{**************************************************************************************************}
{                                                                                                  }
{  LocalRegistry, Version 2.11                                                                     }
{                                                                                                  }
{  The contents of this file are subject to the Y Library Public License Version 1.0 (the          }
{  "License"); you may not use this file except in compliance with the License. You may obtain a   }
{  copy of the License at http://delphi.pjh2.de/                                                   }
{                                                                                                  }
{  Software distributed under the License is distributed on an "AS IS" basis, WITHOUT WARRANTY OF  }
{  ANY KIND, either express or implied. See the License for the specific language governing        }
{  rights and limitations under the License.                                                       }
{                                                                                                  }
{  The Original Code is: LocalRegistry.pas.                                                        }
{  The Initial Developer of the Original Code is Peter J. Haas (libs@pjh2.de). Portions created    }
{  by Peter J. Haas are Copyright (C) 1998-2005 Peter J. Haas. All Rights Reserved.                }
{                                                                                                  }
{  Contributor(s):                                                                                 }
{                                                                                                  }
{  You may retrieve the latest version of this file at the homepage of Peter J. Haas, located at   }
{  http://delphi.pjh2.de/                                                                          }
{                                                                                                  }
{**************************************************************************************************}

// Hinweis: Diese Bibliothek wird von mir nicht weiterentwickelt

// For history see end of file

unit LocalRegistry;

{$ALIGN ON, $BOOLEVAL OFF, $LONGSTRINGS ON, $IOCHECKS ON, $WRITEABLECONST OFF}
{$OVERFLOWCHECKS OFF, $RANGECHECKS OFF, $TYPEDADDRESS ON, $MINENUMSIZE 1}


{ Ist folgendes Symbol definiert, wird zustzlicher Code erzeugt, der es
  ermglicht Dateien zu lesen, die mit den Versionen 2.0x erzeugt wurden. Beim
  Schreiben wird grundstzlich die neue Checksumme erzeugt (PKZIP-kompatibel).
  In neuen Projekten ist die Definition des Symbols nicht notwendig. }

{$define LocalRegistry_ReadOldCRC}

interface
uses
  Windows, SysUtils, Classes {$ifdef debug}, Dialogs {$endif};

type
  TLocalRegistryDataType = (rdtKey,rdtString,rdtFloat,rdtInteger, rdtBoolean,rdtBinary);

{ Bedeutung der Felder
  FName      : Bezeichner fr Wert
  FDataType  : Datentyp

  bei interner Speicherung (max. 10 Byte) :
  FData      : Beginn des Datenbereiches
  FShortSize : Gre der Daten

  bei externer Speicherung (grer 10 Byte) :
  FData      : Pointer auf Daten
  FDataSize  : Gre der Daten
  FShortSize : immer 0 }

  // Listeneintrag   { 16 Byte }
  PLocalRegistryItem = ^TLocalRegistryItem;
  TLocalRegistryItem = packed record
    FName      : String;                  { 4 Byte }
    FData      : Pointer;                 { 4 Byte }
    FDataSize  : Integer;                 { 4 Byte }
    FDataFill1 : Byte;                    { 1 Byte }
    FDataFill2 : Byte;                    { 1 Byte }
    FShortSize : Byte;                    { 1 Byte }
    FDataType  : TLocalRegistryDataType;  { 1 Byte }
  end;

  PLocalRegistryItemList = ^TLocalRegistryItemList;
  TLocalRegistryItemList = array[0..2047] of TLocalRegistryItem;

  // Erweiterung von TMemoryStream
  TLocalMemoryStream = class(TMemoryStream)
  private
    function GetPositionPtr:PByte;
    function GetRestSize:Integer;
  public
    function ReadOrdinal:Integer;
    function ReadInteger:Integer;
    function ReadWord:Word;
    function ReadByte:Byte;
    function ReadString:String;
    procedure WriteOrdinal(AValue:Integer);
    procedure WriteInteger(AValue:Integer);
    procedure WriteWord(AValue:Word);
    procedure WriteByte(AValue:Byte);
    procedure WriteString(AValue:String);
    procedure WriteBinary(const AData; ASize:Integer);
    property PositionPtr:PByte read GetPositionPtr;
    property RestSize:Integer read GetRestSize;
  end;

  // Liste von Eintragungen
  TLocalRegistryList = class(TObject)
  private
    FList      : PLocalRegistryItemList;
    FCount     : Integer;
    FCapacity  : Integer;
    FOwnerList : TLocalRegistryList;
    procedure SetCapacity(ANewCapacity:Integer);
    procedure Grow;
    procedure FreeItem(AIndex:Integer);
    procedure InsertItem(AIndex:Integer; const AName:String;
                         ADataType:TLocalRegistryDataType);
    procedure DeleteItem(AIndex:Integer);
    function Find(const AName:String; var AIndex:Integer):Boolean;
    function OpenOrCreateItem(const AName:String; CanCreate:Boolean;
                              ADataType:TLocalRegistryDataType):Integer;

    // Fehlerbehandlung
    procedure CheckIndex(AIndex:Integer);

    // Lesen und Schreiben, Eigenschaften von Werten
    function GetItemName(Index:Integer):String;
    function GetItemText(Index:Integer):String;
    function GetDataType(Index:Integer):TLocalRegistryDataType;
    function GetDataSize(Index:Integer):Integer;
    // abhngige Werte
    procedure SetAsString (Index:Integer; const Value:String  );
    procedure SetAsFloat  (Index:Integer;       Value:Extended);
    procedure SetAsInteger(Index:Integer;       Value:Integer );
    procedure SetAsBool   (Index:Integer;       Value:LongBool );
    function GetAsString (Index:Integer):String  ;
    function GetAsFloat  (Index:Integer):Extended;
    function GetAsInteger(Index:Integer):Integer ;
    function GetAsBool   (Index:Integer):LongBool ;
  public
    constructor Create(AOwnerList:TLocalRegistryList);
    destructor Destroy; override;
    procedure LoadFromStream(AStream:TLocalMemoryStream);
    procedure SaveToStream(AStream:TLocalMemoryStream);
    function IndexOf(const AName:String): Integer;
    function Add(const AName:String; ADataType:TLocalRegistryDataType):Integer;
    function Rename(const AOldName,ANewName:String):Boolean;
    function DeleteByName(const AName:String; AIsKey:Boolean):Boolean;
    procedure Clear;
    function  AddKey(const AName:String):TLocalRegistryList;
    function  GetKey(const AName:String):TLocalRegistryList;

    procedure GetItems(AList:TStrings);
    procedure GetTexts(AList:TStrings);
    procedure GetKeys(AList:TStrings);
    procedure GetKeyNames(AList:TStrings);
    procedure GetValueNames(AList:TStrings);
    function GetValueCount:Integer;
    function GetKeyCount:Integer;

    // Lesen und Schreiben, Eigenschaften von Werten
    procedure ReadItem (Index:Integer; var ABuffer; ABufSize:Integer);
    procedure WriteItem(Index:Integer; var ABuffer; ABufSize:Integer);

    property OwnerList:TLocalRegistryList read FOwnerList;
    property Count:Integer read FCount;
    // Lesen und Schreiben, Eigenschaften von Werten
    property ItemName[Index:Integer]:String read GetItemName;
    property ItemText[Index:Integer]:String read GetItemText;
    property DataTypes[Index:Integer]:TLocalRegistryDataType read GetDataType;
    property DataSize [Index:Integer]:Integer  read GetDataSize;
    // abhngige Werte
    property AsString [Index:Integer]:String   read GetAsString  write SetAsString;
    property AsFloat  [Index:Integer]:Extended read GetAsFloat   write SetAsFloat;
    property AsInteger[Index:Integer]:Integer  read GetAsInteger write SetAsInteger;
    property AsBool   [Index:Integer]:LongBool read GetAsBool    write SetAsBool;
  end;

  // eigentliches Registryobject
  TLocalRegistry = class(TObject)
  protected
    FRootList    : TLocalRegistryList; { enthlt die Liste }
    FCurrentKey  : String;             { Pfad des aktuellen Zweigs }
    FCurrentList : TLocalRegistryList; { Zeiger auf den aktuellen Zweig }
    procedure CheckCurrentList;
  protected
    FFileType    : String;             { Dateityp (incl. Punkt) }
    FFileBType   : String;             { Dateityp Backupfile }
    FFileDType   : String;             { Dateityp defekte Datei }
    FFileID      : String;             { Kennung }
    procedure InitParams; dynamic;
    procedure DecryptData(ABuffer:PByte; ASize:Integer); dynamic;
    procedure EncryptData(ABuffer:PByte; ASize:Integer); dynamic;
    function CheckIDVersion(m:TLocalMemoryStream):Boolean; dynamic;
  public
    // Nach dem Create wird InitParams aufgerufen
    // die Datei wird nicht automatisch geladen
    constructor Create;
    destructor Destroy; override;
    // Datenzugriff
    function ValueExists(const AValue:String):Boolean;
    procedure ReadBinaryData (const AValue:String; var AData; ADataSize:Integer);
    procedure WriteBinaryData(const AValue:String; var AData; ADataSize:Integer);
    function  ReadString (const AValue:String):String;
    procedure WriteString(const AValue, AData:String);
    function  ReadInteger (const AValue:String):Integer;
    procedure WriteInteger(const AValue:String; AData:Integer);
    function  ReadBool (const AValue:String):LongBool;
    procedure WriteBool(const AValue:String; AData:LongBool);
    function  ReadFloat (const AValue:String):Extended;
    procedure WriteFloat(const AValue:String; AData:Extended);
    function  DeleteValue(const AValue:string):Boolean;
    // Schlsselzugriff
    function  OpenKey(const AKey:String; ACanCreate:Boolean):Boolean;
    function DeleteKey(const AKey: string):Boolean;
    procedure CloseKey;
    procedure GetKeyNames(AList:TStrings);
    procedure GetValueNames(AList:TStrings);
    // Gesamtzugriff
    procedure Clear;
    // Laden von Datei, gibt True zurck, wenn Laden erfolgreich war.
    function  LoadFromFile(const AFilename:String):Boolean;
    // Abspeichern als Datei
    procedure SaveToFile(const AFilename:String);
    property RootList:TLocalRegistryList read FRootList;
    property FileType:String read FFileType;
  end;

  // Exception
  ELocalRegistryError = class(Exception)
  public
    ErrorCode : Integer;
    Ident     : String;
  end;

const
  //Fehlercodes
  lrecType             =  1;  { Typ fehlerhaft            }
  lrecFaultSize        =  2;  { Falsche Datengre fr %s }
  lrecKeyNotExists     =  3;  { Key existiert nicht       }
  lrecUnknownDataType  =  4;  { unbekannter Datentyp      }
  lrecNoList           =  5;  { List nicht erzeugt        }
  lrecBufferSize       =  6;  { Puffer zu klein           }
  lrecIndexRange       =  7;  { Index auerhalb           }
  lrecNoKey            =  8;  { Kein Key                  }
  lrecValueNotExists   =  9;  { %s nicht vorhanden        }
  lrecValueXExists     = 10;  { %s existiert bereits      }

implementation
uses
  CRC32;

const
  CurrentFileVersion = $0210;    { aktuelle Datei-Version 2.10 }

{$ifdef Ver90}
const
{$else}
resourcestring
{$endif}
  // Fehlermeldungen
  SlreFaultType        = 'Typ fehlerhaft';
  SlreFaultSize        = 'Falsche Datengre fr %s';
  SlreKeyNotExists     = 'Key existiert nicht';
  SlreUnknownDataType  = 'unbekannter Datentyp';
  SlreNoList           = 'List nicht erzeugt';
  SlreBufferSize       = 'Puffer zu klein';
  SlreIndexRange       = 'Index auerhalb';
  SlreNoKey            = 'Kein Key';
  SlreValueNotExists   = '%s nicht vorhanden';
  SlreValueXExists     = '%s existiert bereits';
  // unbekannter Fehler
  SlreUnknown          = 'unbekannter Fehler %d';
  // Tabellenberschriften
  SlrtName             = 'Name';
  SlrtValue            = 'Wert';

function CreateError(AErrorCode:Integer; const AIdent:String):ELocalRegistryError;
const
  ErrorMap : array[1..10] of String = (
    SlreFaultType       ,
    SlreFaultSize       ,
    SlreKeyNotExists    ,
    SlreUnknownDataType ,
    SlreNoList          ,
    SlreBufferSize      ,
    SlreIndexRange      ,
    SlreNoKey           ,
    SlreValueNotExists  ,
    SlreValueXExists    );
begin
  case AErrorCode of
    1,3,4..8 : Result:=ELocalRegistryError.Create(ErrorMap[AErrorCode]);
    2,9..10  : Result:=ELocalRegistryError.CreateFmt(ErrorMap[AErrorCode],[AIdent]);
  else
    Result:=ELocalRegistryError.CreateFmt(SlreUnknown,[AErrorCode]);
  end;
  Result.ErrorCode:=AErrorCode;
  Result.Ident:=AIdent;
end;

function TLocalMemoryStream.GetPositionPtr:PByte;
begin
  Result:=Memory;
  Inc(Result,Position);
end;

function TLocalMemoryStream.GetRestSize:Integer;
begin
  Result:=Size-Position;
end;

function TLocalMemoryStream.ReadOrdinal:Integer;
var
  i : ShortInt;  { Lnge der Daten }
begin
  ReadBuffer(i,SizeOf(i));
  case i of
    -3..-1 : begin
      Result:=-1;
      ReadBuffer(Result,-i);
    end;
    0 : Result:=0;
    1..4 : begin
      Result:=0;
      ReadBuffer(Result,i);
    end;
  else
    Seek(-1,soFromCurrent);
    ReadBuffer(Result,SizeOf(Result));
  end;
end;

function TLocalMemoryStream.ReadInteger:Integer;
begin
  ReadBuffer(Result,SizeOf(Result));
end;

function TLocalMemoryStream.ReadWord:Word;
begin
  ReadBuffer(Result,SizeOf(Result));
end;

function TLocalMemoryStream.ReadByte:Byte;
begin
  ReadBuffer(Result,SizeOf(Result));
end;

function TLocalMemoryStream.ReadString:String;
var
  i : Integer;
begin
  i:=ReadOrdinal;
  if i=0 then
    Result:=''
  else begin
    SetLength(Result,i);
    ReadBuffer(Result[1],i);
    Result[i+1]:=#0;
  end;
end;

// Schreibt Integerwert mglichst effektiv in den Stream
procedure TLocalMemoryStream.WriteOrdinal(AValue:Integer);
var
  Count  : ShortInt;
  Mask,i : Integer;
begin
  // Anzahl der bentigten Bytes ermitteln
  Count:=4;
  Mask:=Integer($FF000000);    { mit hchstwertigstem Byte beginnen }
  i:=AValue; if i<0 then i:=i xor -1;
  while ((i and Mask)=0) and (Count>0) do begin
    Mask:=Mask shr 8;          { nchstes Byte }
    Dec(Count);                { Anzahl verringern }
  end;
  if (AValue<0) then begin
    if Count=0 then Count:=1;
    if (Count<4) then Count:=-Count; { negative Zahlen kennzeichnen }
  end;
  // Bei 4 Bytes Count nur schreiben, wenn nicht eindeutig
  if (Count<>4) or ((AValue and $FF) in [$00..$04,$FD..$FF]) then
    WriteBuffer(Count,SizeOf(Count));
  // Wert schreiben
  if Count<>0 then WriteBuffer(AValue,Abs(Count));
end;

procedure TLocalMemoryStream.WriteInteger(AValue:Integer);
begin
  WriteBuffer(AValue,SizeOf(AValue));
end;

procedure TLocalMemoryStream.WriteWord(AValue:Word);
begin
  WriteBuffer(AValue,SizeOf(AValue));
end;

procedure TLocalMemoryStream.WriteByte(AValue:Byte);
begin
  WriteBuffer(AValue,SizeOf(AValue));
end;

procedure TLocalMemoryStream.WriteString(AValue:String);
var
  i : Integer;
begin
  i:=Length(AValue);
  WriteOrdinal(i);
  if i>0 then WriteBuffer(AValue[1],i);
end;

procedure TLocalMemoryStream.WriteBinary(const AData; ASize:Integer);
begin
  WriteOrdinal(ASize);
  if ASize>0 then WriteBuffer(AData,ASize);
end;

constructor TLocalRegistryList.Create(AOwnerList:TLocalRegistryList);
begin
  inherited Create;
  FOwnerList:=AOwnerList;
end;

destructor TLocalRegistryList.Destroy;
begin
  Clear;
  inherited Destroy;
end;

procedure TLocalRegistryList.Grow;
var
  Delta : Integer;
begin
  if FCapacity>64 then Delta:=FCapacity div 4
  else if FCapacity>8 then Delta:=16
  else Delta:=4;
  SetCapacity(FCapacity+Delta);
end;

procedure TLocalRegistryList.SetCapacity(ANewCapacity:Integer);
begin
  ReAllocMem(FList,ANewCapacity*SizeOf(TLocalRegistryItem));
  FCapacity:=ANewCapacity;
end;

procedure TLocalRegistryList.InsertItem(AIndex:Integer; const AName:String;
                                     ADataType:TLocalRegistryDataType);
begin
  if FCount=FCapacity then Grow;
  if AIndex<FCount then
    System.Move(FList^[AIndex],FList^[AIndex+1],
                (FCount-AIndex)*SizeOf(TLocalRegistryItem));
  FillChar(FList^[AIndex],SizeOf(TLocalRegistryItem),0);
  with FList^[AIndex] do begin
    FName:=AName;
    FDataType:=ADataType;
  end;
  Inc(FCount);
end;

procedure TLocalRegistryList.DeleteItem(AIndex:Integer);
begin
  CheckIndex(AIndex);
  FreeItem(AIndex);
  Finalize(FList^[AIndex]);
  Dec(FCount);
  if AIndex<FCount then
    System.Move(FList^[AIndex+1],FList^[AIndex],
                (FCount-AIndex)*SizeOf(TLocalRegistryItem));
end;

function TLocalRegistryList.IndexOf(const AName:String): Integer;
begin
  if Not Find(AName,Result) then Result:=-1;
end;

function TLocalRegistryList.Find(const AName:String; var AIndex:Integer):Boolean;
var
  L,H,I,C : Integer;
begin
  Result:=False;
  L:=0;
  H:=FCount-1;
  while L<=H do begin
    I:=(L+H) shr 1;
    C:=AnsiCompareText(FList^[i].FName,AName);
    if C<0 then
      L:=I+1
    else begin
      H:=I-1;
      if C=0 then begin
        Result:=True;
        L:=I;
      end;
    end;
  end;
  AIndex:=L;
end;

function TLocalRegistryList.Add(const AName:String; ADataType:TLocalRegistryDataType):Integer;
begin
  if Find(AName,Result) then raise CreateError(lrecValueXExists,AName);
  InsertItem(Result,AName,ADataType);
  case ADataType of
    rdtKey : FList^[Result].FData:=TLocalRegistryList.Create(Self);
  end;
end;

function TLocalRegistryList.Rename(const AOldName,ANewName:String):Boolean;
var
  i : Integer;
  xItem : TLocalRegistryItem;
begin
  Result:=Find(AOldName,i);  { Eintrag suchen }
  if Result then begin
    if AnsiCompareText(AOldName,ANewName)=0 then begin { Nur Schreibweise gendert }
      FList^[i].FName:=ANewName;
    end
    else begin
      Result:=IndexOf(ANewName)<0;
      if Result then begin  { neuer Name nicht vorhanden }
        xItem:=FList^[i];
        xItem.FName:=ANewName;
        DeleteItem(i);
        Find(ANewName,i);
        InsertItem(i,ANewName,xItem.FDataType);
        FList^[i]:=xItem;
      end;
    end;
  end;
end;

function TLocalRegistryList.DeleteByName(const AName:String; AIsKey:Boolean):Boolean;
var
  i : Integer;
begin
  Result:=Find(AName,i);
  if Result then begin
    Result:=(FList^[i].FDataType<>rdtKey) xor AIsKey;
    if Result then DeleteItem(i);
  end;
end;

procedure TLocalRegistryList.FreeItem(AIndex:Integer);
begin
  with FList^[AIndex] do begin
    case FDataType of
      rdtKey    : if FData<>Nil then TLocalRegistryList(FData).Free;
    else
      if (FShortSize=0) and (FDataSize>0) then
        if FData<>Nil then FreeMem(FData,FDataSize);
    end;
  end;
end;

procedure TLocalRegistryList.Clear;
var
  i : Integer;
begin
  for i:=0 to FCount-1 do FreeItem(i);
  if FCount>0 then Finalize(FList^[0],FCount);
  FCount:=0;
  SetCapacity(0);
end;

function TLocalRegistryList.OpenOrCreateItem(const AName:String; CanCreate:Boolean;
                                          ADataType:TLocalRegistryDataType):Integer;
begin
  Result:=IndexOf(AName);
  if Result<0 then begin
    if CanCreate then
      Result:=Add(AName,ADataType)
    else
      raise CreateError(lrecValueNotExists,AName);
  end;
end;

function  TLocalRegistryList.AddKey(const AName:String):TLocalRegistryList;
var
  i : Integer;
begin
  i:=Add(AName,rdtKey);
  Result:=TLocalRegistryList(FList^[i].FData);
end;

function  TLocalRegistryList.GetKey(const AName:String):TLocalRegistryList;
var
  i : Integer;
begin
  i:=OpenOrCreateItem(AName,False,rdtKey);
  with FList^[i] do begin
    if FDataType<>rdtKey then raise CreateError(lrecNoKey,'');
    Result:=TLocalRegistryList(FData);
  end;
end;

procedure TLocalRegistryList.CheckIndex(AIndex:Integer);
begin
  if (AIndex<0) or (AIndex>=FCount) then raise CreateError(lrecIndexRange,'');
end;

// Routinen
procedure TLocalRegistryList.SetAsString (Index:Integer; const Value:String);
begin
  WriteItem(Index,PChar(Value)^,Length(Value)+1);
end;

procedure TLocalRegistryList.SetAsFloat  (Index:Integer;       Value:Extended);
begin
  WriteItem(Index,Value,SizeOf(Value));
end;

procedure TLocalRegistryList.SetAsInteger(Index:Integer;       Value:Integer );
begin
  WriteItem(Index,Value,SizeOf(Value));
end;

procedure TLocalRegistryList.SetAsBool   (Index:Integer;       Value:LongBool);
begin
  WriteItem(Index,Value,SizeOf(Value));
end;

function TLocalRegistryList.GetAsString (Index:Integer):String  ;
var
  i : Integer;
begin
  i:=GetDataSize(Index);
  if i>1 then begin
    SetLength(Result,i-1);
    ReadItem(Index,PChar(Result)^,i);
  end
  else
    Result:='';
end;

function TLocalRegistryList.GetAsFloat  (Index:Integer):Extended;
begin
  ReadItem(Index,Result,SizeOf(Result));
end;

function TLocalRegistryList.GetAsInteger(Index:Integer):Integer ;
begin
  ReadItem(Index,Result,SizeOf(Result));
end;

function TLocalRegistryList.GetAsBool   (Index:Integer):LongBool ;
begin
  ReadItem(Index,Result,SizeOf(Result));
end;

// physischer Zugriff auf die Daten
procedure TLocalRegistryList.ReadItem(Index:Integer; var ABuffer; ABufSize:Integer);
var
  i : Integer;
begin
  CheckIndex(Index);
  with FList^[Index] do begin
    if FDataType=rdtKey then raise CreateError(lrecType,'');
    i:=FShortSize; if i=0 then i:=FDataSize;
    if i=0 then Exit;
    if i>ABufSize then raise CreateError(lrecBufferSize,'');
    if FShortSize=0 then { externe Speicherung }
      System.Move(FData^,ABuffer,FDataSize)
    else
      System.Move(FData,ABuffer,FShortSize);
  end;
end;

procedure TLocalRegistryList.WriteItem(Index:Integer; var ABuffer; ABufSize:Integer);
begin
  CheckIndex(Index);
  with FList^[Index] do begin
    case FDataType of
      rdtString, rdtBinary : ;
      rdtFloat   : if ABufSize<>10 then raise CreateError(lrecFaultSize,'Float');
      rdtInteger : if ABufSize<> 4 then raise CreateError(lrecFaultSize,'Integer');
      rdtBoolean : if ABufSize<> 4 then raise CreateError(lrecFaultSize,'Boolean');
    else
      raise CreateError(lrecType,'');
    end;
    if ABufSize>10 then begin { neu = extern }
      if FShortSize>0 then   { internen Puffer lschen }
        FillChar(FData,11,0);
      if FDataSize<>ABufSize then begin
        ReAllocMem(FData,ABufSize);
        FDataSize:=ABufSize;
      end;
      System.Move(ABuffer,FData^,ABufSize);
    end
    else begin { neu = intern }
      if (FShortSize=0) and (FDataSize>0) then  { Puffer freigeben }
        ReAllocMem(FData,0);
      FillChar(FData,10,0);
      FShortSize:=ABufSize;
      if ABufSize>0 then System.Move(ABuffer,FData,ABufSize);
    end;
  end;
end;

function TLocalRegistryList.GetItemName (Index:Integer):String;
begin
  CheckIndex(Index);
  Result:=FList^[Index].FName;
end;

function BufferToHex(var AData; ACount:Integer):String;
var
  i : Integer;
begin
  Result:='';
  for i:=0 to ACount-1 do
    Result:=Format('%s%.2x ',[Result,TByteArray(AData)[i]]);
  Delete(Result,Length(Result),1);
end;

function TLocalRegistryList.GetItemText(Index:Integer):String;
begin
  CheckIndex(Index);
  with FList^[Index] do begin
    case FDataType of
      rdtString  : Result:=GetAsString(Index);
      rdtFloat   : Result:=FloatToStr(GetAsFloat(Index));
      rdtInteger : Result:=Format('$%.8x (%0:d)',[GetAsInteger(Index)]);
      rdtBoolean : if GetAsBool(Index) then Result:='True' else Result:='False';
      rdtBinary  : begin
        if FShortSize>0 then     { interne Speicherung }
          Result:=BufferToHex(FData,FShortSize)
        else if FDataSize>0 then
          Result:=BufferToHex(FData^,FDataSize)
        else
          Result:='';
      end;
    else
      Result:='';
    end;
  end;
end;

function TLocalRegistryList.GetDataType(Index:Integer):TLocalRegistryDataType;
begin
  CheckIndex(Index);
  with FList^[Index] do Result:=FDataType;
end;

function TLocalRegistryList.GetDataSize(Index:Integer):Integer;
begin
  CheckIndex(Index);
  with FList^[Index] do begin
    Result:=FShortSize;
    if Result=0 then Result:=FDataSize;
  end;
end;

procedure TLocalRegistryList.GetItems (AList:TStrings);
var
  i : Integer;
begin
  with AList do begin
    BeginUpdate;
    try
      Clear;
      Add(SlrtName);
      for i:=0 to FCount-1 do with FList^[i] do
        if FDataType<>rdtKey then AList.Add(FName);
    finally
      EndUpdate;
    end;
  end;
end;

procedure TLocalRegistryList.GetTexts(AList:TStrings);
var
  i : Integer;
begin
  with AList do begin
    BeginUpdate;
    try
      Clear;
      Add(SlrtValue);
      for i:=0 to FCount-1 do with FList^[i] do
        if FDataType<>rdtKey then
          AList.Add(GetItemText(i));
    finally
      EndUpdate;
    end;
  end;
end;

procedure TLocalRegistryList.GetKeys(AList:TStrings);
var
  i : Integer;
begin
  with AList do begin
    BeginUpdate;
    try
      Clear;
      for i:=0 to FCount-1 do with FList^[i] do
        if FDataType=rdtKey then AddObject(FName,TObject(FData));
    finally
      EndUpdate;
    end;
  end;
end;

procedure TLocalRegistryList.GetValueNames(AList:TStrings);
var
  i : Integer;
begin
  with AList do begin
    BeginUpdate;
    try
      Clear;
      for i:=0 to FCount-1 do with FList^[i] do
        if FDataType<>rdtKey then AList.Add(FName);
    finally
      EndUpdate;
    end;
  end;
end;

procedure TLocalRegistryList.GetKeyNames(AList:TStrings);
var
  i : Integer;
begin
  with AList do begin
    BeginUpdate;
    try
      Clear;
      for i:=0 to FCount-1 do with FList^[i] do
        if FDataType=rdtKey then Add(FName);
    finally
      EndUpdate;
    end;
  end;
end;

function TLocalRegistryList.GetValueCount:Integer;
var
  i : Integer;
begin
  Result:=0;
  for i:=0 to FCount-1 do
    if FList^[i].FDataType<>rdtKey then Inc(Result);
end;

function TLocalRegistryList.GetKeyCount:Integer;
var
  i : Integer;
begin
  Result:=0;
  for i:=0 to FCount-1 do
    if FList^[i].FDataType=rdtKey then Inc(Result);
end;

procedure TLocalRegistryList.LoadFromStream(AStream:TLocalMemoryStream);
var
  i         : Integer;
  xCount    : Integer;
  xIndex    : Integer;
  xName     : String;
  xDataType : TLocalRegistryDataType;
  xDataSize : Integer;
begin
  xCount:=AStream.ReadOrdinal;
  SetCapacity(xCount);
  for i:=0 to xCount-1 do begin
    xName    :=AStream.ReadString;
    xDataType:=TLocalRegistryDataType(AStream.ReadByte);
    xIndex:=Add(xName,xDataType);
    with FList^[xIndex] do begin
      case FDataType of
        rdtKey         : begin
          if FData=Nil then raise CreateError(lrecNoList,'');
          TLocalRegistryList(FData).LoadFromStream(AStream);
        end;
        rdtString,rdtFloat,rdtInteger,rdtBoolean,rdtBinary : begin
          xDataSize:=AStream.ReadOrdinal;
          if xDataSize>0 then begin
            if xDataSize>10 then begin
              GetMem(FData,xDataSize);
              AStream.ReadBuffer(FData^,xDataSize);
              FDataSize:=xDataSize;
            end
            else begin
              AStream.ReadBuffer(FData,xDataSize);
              FShortSize:=xDataSize;
            end;
          end;
        end;
      else
        raise CreateError(lrecUnknownDataType,'');
      end;
    end;
  end;
end;

procedure TLocalRegistryList.SaveToStream(AStream:TLocalMemoryStream);
var
  i : Integer;
begin
  AStream.WriteOrdinal(FCount);
  for i:=0 to FCount-1 do with FList^[i] do begin
    AStream.WriteString(FName);
    AStream.WriteByte(Ord(FDataType));
    case FDataType of
      rdtKey         : begin
        if FData=Nil then
          AStream.WriteOrdinal(0)   { Anzahl = 0 }
        else
          TLocalRegistryList(FData).SaveToStream(AStream);
      end;
      rdtString,rdtFloat,rdtInteger,rdtBoolean,rdtBinary : begin
        if FShortSize>0 then begin
          AStream.WriteBinary(FData,FShortSize);
        end
        else if FDataSize>0 then begin
          AStream.WriteBinary(FData^,FDataSize);
        end
        else begin
          AStream.WriteOrdinal(0);
        end;
      end;
    else
      raise CreateError(lrecUnknownDataType,'');
    end;
  end;
end;

{ ******************************************************************* }

constructor TLocalRegistry.Create;
begin
  inherited Create;
  FRootList:=TLocalRegistryList.Create(Nil);
  CloseKey;
  InitParams;
end;

procedure TLocalRegistry.InitParams;
begin
  FFileType :='.lrg';           { Dateityp (incl. Punkt) }
  FFileBType:='.~lr';           { Dateityp Backupfile }
  FFileDType:='Defekt.~lr';     { Dateityp defekte Datei }
  FFileID   :='LocalRegistry';  { Kennung }
end;

destructor TLocalRegistry.Destroy;
begin
  FRootList.Free;
  inherited Destroy;
end;

function TLocalRegistry.CheckIDVersion(m:TLocalMemoryStream):Boolean;
var
  ID : array[0..32] of Char;
  Version : Word;
  Chksum : DWord;
begin
  // Kennung lesen
  m.ReadBuffer(ID,Length(FFileID));
  ID[Length(FFileID)]:=#0;
  if (StrLComp(ID,PChar(FFileID),Length(FFileID))=0) then begin
    // Version lesen
    m.ReadBuffer(Version,SizeOf(Version));
    {$ifdef LocalRegistry_ReadOldCRC}
    case Version of
      $0200, CurrentFileVersion : begin
        m.ReadBuffer(Chksum,SizeOf(Chksum));
        if Version=$0200 then begin
          Result:=DWord(CRC32R(0,m.PositionPtr^,m.RestSize))=Chksum;
        end
        else begin
          Result:=CalcCRC32PKZIP(0,m.PositionPtr,m.RestSize)=Chksum;
        end;
      end;
    else
      Result:=False;
    end;
    {$else}
    if Version=CurrentFileVersion then begin
      m.ReadBuffer(Chksum,SizeOf(Chksum));
      Result:=CRC32R(0,m.PositionPtr^,m.RestSize)=Chksum;
    end
    else Result:=False;
    {$endif}
  end
  else Result:=False;
end;

procedure TLocalRegistry.DecryptData(ABuffer:PByte; ASize:Integer);
begin
end;

procedure TLocalRegistry.EncryptData(ABuffer:PByte; ASize:Integer);
begin
end;

function TLocalRegistry.LoadFromFile(const AFilename:String):Boolean;
var
  m : TLocalMemoryStream;
  f,fb,fd : String;

function LoadAndDecryptFile:Boolean;
begin
  m.LoadFromFile(f);
  // Auswerten der Kennung (Konvertierung alte Version)
  Result:=CheckIDVersion(m);
  if Result then begin
    // Entschlsseln
    DecryptData(m.PositionPtr,m.RestSize);
  end;
end;

begin
  // Dateinamen mit Endungen erzeugen
  f:=ChangeFileExt(AFilename,'');
  fb:=f+FFileBType;
  fd:=f+FFileDType;
  f:=f+FFileType;
  // prfen, ob Datei existiert
  Result:=FileExists(f);
  {$ifdef debug} ShowMessage('Open:'+f); {$endif}
  if Result then begin
    m:=TLocalMemoryStream.Create;
    try
      Result:=LoadAndDecryptFile;
      if Not Result then begin
        // Datei beschdigt
        if FileExists(fd) then DeleteFile(fd);    { alte defekte Datei lschen }
        RenameFile(f,fd);                         { aktuelle Datei in defekt umbenennen }
        if FileExists(fb) then RenameFile(fb,f);  { Backupdatei in aktuelle umbenennen }
        if FileExists(f) then Result:=LoadAndDecryptFile;   { zweiter Versuch }
      end;
      if Result then begin
        Clear;  { alte Liste freigeben }
        FRootList.LoadFromStream(m);
      end;
    finally
      m.Free;
    end;
  end;
end;

procedure TLocalRegistry.SaveToFile(const AFilename:String);
var
  f, fb : String;
  m : TLocalMemoryStream;
  Version : Word;
  ChkSum : Integer;
begin
  // Dateinamen mit Endungen erzeugen
  f:=ChangeFileExt(AFilename,'');
  if Length(f)=0 then Exit;
  fb:=f+FFileBType;
  f:=f+FFileType;
  DeleteFile(fb);                           { altes Backupfile lschen }
  if FileExists(f) then RenameFile(f,fb);   { alte Datei als Backupfile umbenennen }
  m:=TLocalMemoryStream.Create;
  try
    // Kennung schreiben
    m.WriteBuffer(FFileID[1],Length(FFileID));
    Version:=CurrentFileVersion;
    m.WriteBuffer(Version,SizeOf(Version));
    ChkSum:=0;
    m.WriteBuffer(ChkSum,SizeOf(ChkSum));   { als Platzhalter fr Checksumme schreiben }
    ChkSum:=m.Position;                     { Position Datenbeginn merken }
    // Daten schreiben
    FRootList.SaveToStream(m);
    // Checksumme berechnen und eintragen
    m.Seek(ChkSum,soFromBeginning);              { Stream auf Datenbeginn setzen }
    ChkSum:=CalcCRC32PKZIP(0,m.PositionPtr,m.RestSize); { Checksumme berechnen }
    m.Seek(-SizeOf(ChkSum),soFromCurrent);       { Stream auf Checksumme setzen }
    m.WriteBuffer(ChkSum,SizeOf(ChkSum));        { und schreiben }
    EncryptData(m.PositionPtr,m.RestSize);
    m.SaveToFile(f);
  finally
    m.Free;
  end;
end;

// Schlsselzugriff
function TLocalRegistry.OpenKey(const AKey:String; ACanCreate:Boolean):Boolean;
var
  i : Integer;
  s : String;
  Path : String;

function FetchKey:String;
var
  i : Integer;
begin
  i:=Pos('\',Path);
  if i<1 then i:=Length(Path)+1;
  Result:=Copy(Path,1,i-1);
  Delete(Path,1,i);
end;

begin
  Result:=Length(AKey)>0;
  if Result then begin
    Path:=AKey;
    if Path[1]='\' then begin { absolute Pfadangabe }
      FCurrentList:=FRootList;
      Delete(Path,1,1);
    end;
    while Length(Path)>0 do begin
      s:=FetchKey;
      i:=FCurrentList.IndexOf(s);
      if i<0 then begin   { nicht vorhanden }
        if ACanCreate then begin
          FCurrentList:=FCurrentList.AddKey(s);
        end
        else begin
          Result:=False;
          Break;
        end;
      end
      else begin { Key vorhanden }
        FCurrentList:=FCurrentList.GetKey(s);
      end;
    end;
  end
  else begin
    CloseKey;
  end;
end;

procedure TLocalRegistry.CloseKey;
begin
  FCurrentKey:='';
  FCurrentList:=FRootList;
end;

procedure TLocalRegistry.CheckCurrentList;
begin
  if FCurrentList=Nil then raise CreateError(lrecKeyNotExists,'');
end;

// Datenzugriff
function TLocalRegistry.ValueExists(const AValue:String):Boolean;
begin
  Result:=FCurrentList<>Nil;
  if Result then Result:=FCurrentList.IndexOf(AValue)>=0;
end;

function  TLocalRegistry.ReadString (const AValue:String):String;
var
  i : Integer;
begin
  CheckCurrentList;
  with FCurrentList do begin
    i:=OpenOrCreateItem(AValue,False,rdtString);
    Result:=AsString[i];
  end;
end;

procedure TLocalRegistry.WriteString(const AValue, AData:String);
var
  i : Integer;
begin
  CheckCurrentList;
  with FCurrentList do begin
    i:=OpenOrCreateItem(AValue,True,rdtString);
    AsString[i]:=AData;
  end;
end;

function  TLocalRegistry.ReadFloat (const AValue:String):Extended;
var
  i : Integer;
begin
  CheckCurrentList;
  with FCurrentList do begin
    i:=OpenOrCreateItem(AValue,False,rdtFloat);
    Result:=AsFloat[i];
  end;
end;

procedure TLocalRegistry.WriteFloat(const AValue:String; AData:Extended);
var
  i : Integer;
begin
  CheckCurrentList;
  with FCurrentList do begin
    i:=OpenOrCreateItem(AValue,True,rdtFloat);
    AsFloat[i]:=AData;
  end;
end;

function  TLocalRegistry.ReadInteger (const AValue:String):Integer;
var
  i : Integer;
begin
  CheckCurrentList;
  with FCurrentList do begin
    i:=OpenOrCreateItem(AValue,False,rdtInteger);
    Result:=AsInteger[i];
  end;
end;

procedure TLocalRegistry.WriteInteger(const AValue:String; AData:Integer);
var
  i : Integer;
begin
  CheckCurrentList;
  with FCurrentList do begin
    i:=OpenOrCreateItem(AValue,True,rdtInteger);
    AsInteger[i]:=AData;
  end;
end;

function  TLocalRegistry.ReadBool (const AValue:String):LongBool;
var
  i : Integer;
begin
  CheckCurrentList;
  with FCurrentList do begin
    i:=OpenOrCreateItem(AValue,False,rdtBoolean);
    Result:=AsBool[i];
  end;
end;

procedure TLocalRegistry.WriteBool(const AValue:String; AData:LongBool);
var
  i : Integer;
begin
  CheckCurrentList;
  with FCurrentList do begin
    i:=OpenOrCreateItem(AValue,True,rdtBoolean);
    AsBool[i]:=AData;
  end;
end;

procedure TLocalRegistry.ReadBinaryData (const AValue:String; var AData; ADataSize:Integer);
var
  i : Integer;
begin
  CheckCurrentList;
  with FCurrentList do begin
    i:=OpenOrCreateItem(AValue,False,rdtBinary);
    ReadItem(i,AData,ADataSize);
  end;
end;

procedure TLocalRegistry.WriteBinaryData(const AValue:String; var AData; ADataSize:Integer);
var
  i : Integer;
begin
  CheckCurrentList;
  with FCurrentList do begin
    i:=OpenOrCreateItem(AValue,True,rdtBinary);
    WriteItem(i,AData,ADataSize);
  end;
end;

procedure TLocalRegistry.Clear;
begin
  FRootList.Clear;
end;

procedure TLocalRegistry.GetKeyNames(AList:TStrings);
begin
  CheckCurrentList;
  FCurrentList.GetKeyNames(AList);
end;

procedure TLocalRegistry.GetValueNames(AList:TStrings);
begin
  CheckCurrentList;
  FCurrentList.GetValueNames(AList);
end;

function TLocalRegistry.DeleteKey(const AKey: string):Boolean;
begin
  CheckCurrentList;
  Result:=FCurrentList.DeleteByName(AKey,True);
end;

function TLocalRegistry.DeleteValue(const AValue:string):Boolean;
begin
  CheckCurrentList;
  Result:=FCurrentList.DeleteByName(AValue,False);
end;

// *******************************************************************************************

//  History:
//   2005-02-19, Peter J. Haas
//    - neue Lizenz
//
//   2000-10-22, Peter J. Haas, Version 2.11
//    - Umstellung der Lizenz auf MPL                                          
//    - Anpassung an Delphi 4/5
//
//   2000-06-26, Peter J. Haas, Version 2.10
//    - Fehler in der CRC-Unit entdeckt und korrigiert
//    - Korrekte CRC (nach PKZIP) und automatische Konvertierung der alten Version eingebaut
//
//   2000-05-25, Peter J. Haas, Version 2.02
//    - DeleteKey, DeleteValue hinzugefgt
//
//   2000-05-06, Peter J. Haas, Version 2.01
//    - GetKeyNames, GetValueNames hinzugefgt
//
//   1999-07-30, Peter J. Haas, Version 2.0 (beta)
//    - erste Verffentlichung
//
//   1998, Peter J. Haas, Version 1.x
//    - keine Verffentlichung

end.
